@@ -1,4 +1,5 @@
 import collections
+import importlib.metadata
 import json
 import logging
 import os
@@ -8,13 +9,16 @@ from queue import Queue
 from subprocess import CalledProcessError
 
 import colorama
-import pkg_resources
 
 import borgmatic.actions.borg
 import borgmatic.actions.break_lock
 import borgmatic.actions.check
 import borgmatic.actions.compact
+import borgmatic.actions.config.bootstrap
+import borgmatic.actions.config.generate
+import borgmatic.actions.config.validate
 import borgmatic.actions.create
+import borgmatic.actions.export_key
 import borgmatic.actions.export_tar
 import borgmatic.actions.extract
 import borgmatic.actions.info
@@ -26,67 +30,84 @@ import borgmatic.actions.restore
 import borgmatic.actions.rinfo
 import borgmatic.actions.rlist
 import borgmatic.actions.transfer
-import borgmatic.commands.completion
+import borgmatic.commands.completion.bash
+import borgmatic.commands.completion.fish
 from borgmatic.borg import umount as borg_umount
 from borgmatic.borg import version as borg_version
 from borgmatic.commands.arguments import parse_arguments
-from borgmatic.config import checks, collect, convert, validate
+from borgmatic.config import checks, collect, validate
 from borgmatic.hooks import command, dispatch, monitor
-from borgmatic.logger import add_custom_log_levels, configure_logging, should_do_markup
+from borgmatic.logger import DISABLED, add_custom_log_levels, configure_logging, should_do_markup
 from borgmatic.signals import configure_signals
 from borgmatic.verbosity import verbosity_to_log_level
 
 logger = logging.getLogger(__name__)
 
-LEGACY_CONFIG_PATH = '/etc/borgmatic/config'
 
-
-def run_configuration(config_filename, config, arguments):
+def get_skip_actions(config, arguments):
     '''
-    Given a config filename, the corresponding parsed config dict, and command-line arguments as a
-    dict from subparser name to a namespace of parsed arguments, execute the defined create, prune,
-    compact, check, and/or other actions.
+    Given a configuration dict and command-line arguments as an argparse.Namespace, return a list of
+    the configured action names to skip. Omit "check" from this list though if "check --force" is
+    part of the command-like arguments.
+    '''
+    skip_actions = config.get('skip_actions', [])
+
+    if 'check' in arguments and arguments['check'].force:
+        return [action for action in skip_actions if action != 'check']
+
+    return skip_actions
+
+
+def run_configuration(config_filename, config, config_paths, arguments):
+    '''
+    Given a config filename, the corresponding parsed config dict, a sequence of loaded
+    configuration paths, and command-line arguments as a dict from subparser name to a namespace of
+    parsed arguments, execute the defined create, prune, compact, check, and/or other actions.
 
     Yield a combination of:
 
       * JSON output strings from successfully executing any actions that produce JSON
       * logging.LogRecord instances containing errors from any actions or backup hooks that fail
     '''
-    (location, storage, retention, consistency, hooks) = (
-        config.get(section_name, {})
-        for section_name in ('location', 'storage', 'retention', 'consistency', 'hooks')
-    )
     global_arguments = arguments['global']
 
-    local_path = location.get('local_path', 'borg')
-    remote_path = location.get('remote_path')
-    retries = storage.get('retries', 0)
-    retry_wait = storage.get('retry_wait', 0)
+    local_path = config.get('local_path', 'borg')
+    remote_path = config.get('remote_path')
+    retries = config.get('retries', 0)
+    retry_wait = config.get('retry_wait', 0)
     encountered_error = None
     error_repository = ''
     using_primary_action = {'create', 'prune', 'compact', 'check'}.intersection(arguments)
     monitoring_log_level = verbosity_to_log_level(global_arguments.monitoring_verbosity)
+    monitoring_hooks_are_activated = using_primary_action and monitoring_log_level != DISABLED
+    skip_actions = get_skip_actions(config, arguments)
+
+    if skip_actions:
+        logger.debug(
+            f"{config_filename}: Skipping {'/'.join(skip_actions)} action{'s' if len(skip_actions) > 1 else ''} due to configured skip_actions"
+        )
 
     try:
-        local_borg_version = borg_version.local_borg_version(storage, local_path)
+        local_borg_version = borg_version.local_borg_version(config, local_path)
+        logger.debug(f'{config_filename}: Borg {local_borg_version}')
     except (OSError, CalledProcessError, ValueError) as error:
         yield from log_error_records(f'{config_filename}: Error getting local Borg version', error)
         return
 
     try:
-        if using_primary_action:
+        if monitoring_hooks_are_activated:
             dispatch.call_hooks(
                 'initialize_monitor',
-                hooks,
+                config,
                 config_filename,
                 monitor.MONITOR_HOOK_NAMES,
                 monitoring_log_level,
                 global_arguments.dry_run,
             )
-        if using_primary_action:
+
             dispatch.call_hooks(
                 'ping_monitor',
-                hooks,
+                config,
                 config_filename,
                 monitor.MONITOR_HOOK_NAMES,
                 monitor.State.START,
@@ -102,25 +123,28 @@ def run_configuration(config_filename, config, arguments):
 
     if not encountered_error:
         repo_queue = Queue()
-        for repo in location['repositories']:
-            repo_queue.put((repo, 0),)
+        for repo in config['repositories']:
+            repo_queue.put(
+                (repo, 0),
+            )
 
         while not repo_queue.empty():
             repository, retry_num = repo_queue.get()
-            logger.debug(f'{repository["path"]}: Running actions for repository')
+            logger.debug(
+                f'{repository.get("label", repository["path"])}: Running actions for repository'
+            )
             timeout = retry_num * retry_wait
             if timeout:
-                logger.warning(f'{config_filename}: Sleeping {timeout}s before next retry')
+                logger.warning(
+                    f'{repository.get("label", repository["path"])}: Sleeping {timeout}s before next retry'
+                )
                 time.sleep(timeout)
             try:
                 yield from run_actions(
                     arguments=arguments,
                     config_filename=config_filename,
-                    location=location,
-                    storage=storage,
-                    retention=retention,
-                    consistency=consistency,
-                    hooks=hooks,
+                    config=config,
+                    config_paths=config_paths,
                     local_path=local_path,
                     remote_path=remote_path,
                     local_borg_version=local_borg_version,
@@ -128,35 +152,38 @@ def run_configuration(config_filename, config, arguments):
                 )
             except (OSError, CalledProcessError, ValueError) as error:
                 if retry_num < retries:
-                    repo_queue.put((repository, retry_num + 1),)
+                    repo_queue.put(
+                        (repository, retry_num + 1),
+                    )
                     tuple(  # Consume the generator so as to trigger logging.
                         log_error_records(
-                            f'{repository["path"]}: Error running actions for repository',
+                            f'{repository.get("label", repository["path"])}: Error running actions for repository',
                             error,
                             levelno=logging.WARNING,
                             log_command_error_output=True,
                         )
                     )
                     logger.warning(
-                        f'{config_filename}: Retrying... attempt {retry_num + 1}/{retries}'
+                        f'{repository.get("label", repository["path"])}: Retrying... attempt {retry_num + 1}/{retries}'
                     )
                     continue
 
                 if command.considered_soft_failure(config_filename, error):
-                    return
+                    break
 
                 yield from log_error_records(
-                    f'{repository["path"]}: Error running actions for repository', error
+                    f'{repository.get("label", repository["path"])}: Error running actions for repository',
+                    error,
                 )
                 encountered_error = error
                 error_repository = repository['path']
 
     try:
-        if using_primary_action:
-            # send logs irrespective of error
+        if monitoring_hooks_are_activated:
+            # Send logs irrespective of error.
             dispatch.call_hooks(
                 'ping_monitor',
-                hooks,
+                config,
                 config_filename,
                 monitor.MONITOR_HOOK_NAMES,
                 monitor.State.LOG,
@@ -164,18 +191,16 @@ def run_configuration(config_filename, config, arguments):
                 global_arguments.dry_run,
             )
     except (OSError, CalledProcessError) as error:
-        if command.considered_soft_failure(config_filename, error):
-            return
-
-        encountered_error = error
-        yield from log_error_records(f'{repository["path"]}: Error pinging monitor', error)
+        if not command.considered_soft_failure(config_filename, error):
+            encountered_error = error
+            yield from log_error_records(f'{repository["path"]}: Error pinging monitor', error)
 
     if not encountered_error:
         try:
-            if using_primary_action:
+            if monitoring_hooks_are_activated:
                 dispatch.call_hooks(
                     'ping_monitor',
-                    hooks,
+                    config,
                     config_filename,
                     monitor.MONITOR_HOOK_NAMES,
                     monitor.State.FINISH,
@@ -184,7 +209,7 @@ def run_configuration(config_filename, config, arguments):
                 )
                 dispatch.call_hooks(
                     'destroy_monitor',
-                    hooks,
+                    config,
                     config_filename,
                     monitor.MONITOR_HOOK_NAMES,
                     monitoring_log_level,
@@ -200,8 +225,8 @@ def run_configuration(config_filename, config, arguments):
     if encountered_error and using_primary_action:
         try:
             command.execute_hook(
-                hooks.get('on_error'),
-                hooks.get('umask'),
+                config.get('on_error'),
+                config.get('umask'),
                 config_filename,
                 'on-error',
                 global_arguments.dry_run,
@@ -211,7 +236,7 @@ def run_configuration(config_filename, config, arguments):
             )
             dispatch.call_hooks(
                 'ping_monitor',
-                hooks,
+                config,
                 config_filename,
                 monitor.MONITOR_HOOK_NAMES,
                 monitor.State.FAIL,
@@ -220,7 +245,7 @@ def run_configuration(config_filename, config, arguments):
             )
             dispatch.call_hooks(
                 'destroy_monitor',
-                hooks,
+                config,
                 config_filename,
                 monitor.MONITOR_HOOK_NAMES,
                 monitoring_log_level,
@@ -237,11 +262,8 @@ def run_actions(
     *,
     arguments,
     config_filename,
-    location,
-    storage,
-    retention,
-    consistency,
-    hooks,
+    config,
+    config_paths,
     local_path,
     remote_path,
     local_borg_version,
@@ -249,9 +271,9 @@ def run_actions(
 ):
     '''
     Given parsed command-line arguments as an argparse.ArgumentParser instance, the configuration
-    filename, several different configuration dicts, local and remote paths to Borg, a local Borg
-    version string, and a repository name, run all actions from the command-line arguments on the
-    given repository.
+    filename, a configuration dict, a sequence of loaded configuration paths, local and remote paths
+    to Borg, a local Borg version string, and a repository name, run all actions from the
+    command-line arguments on the given repository.
 
     Yield JSON output strings from executing any actions that produce JSON.
 
@@ -266,47 +288,47 @@ def run_actions(
     hook_context = {
         'repository': repository_path,
         # Deprecated: For backwards compatibility with borgmatic < 1.6.0.
-        'repositories': ','.join([repo['path'] for repo in location['repositories']]),
+        'repositories': ','.join([repo['path'] for repo in config['repositories']]),
         'log_file': global_arguments.log_file if global_arguments.log_file else '',
     }
+    skip_actions = set(get_skip_actions(config, arguments))
 
     command.execute_hook(
-        hooks.get('before_actions'),
-        hooks.get('umask'),
+        config.get('before_actions'),
+        config.get('umask'),
         config_filename,
         'pre-actions',
         global_arguments.dry_run,
         **hook_context,
     )
 
-    for (action_name, action_arguments) in arguments.items():
-        if action_name == 'rcreate':
+    for action_name, action_arguments in arguments.items():
+        if action_name == 'rcreate' and action_name not in skip_actions:
             borgmatic.actions.rcreate.run_rcreate(
                 repository,
-                storage,
+                config,
                 local_borg_version,
                 action_arguments,
                 global_arguments,
                 local_path,
                 remote_path,
             )
-        elif action_name == 'transfer':
+        elif action_name == 'transfer' and action_name not in skip_actions:
             borgmatic.actions.transfer.run_transfer(
                 repository,
-                storage,
+                config,
                 local_borg_version,
                 action_arguments,
                 global_arguments,
                 local_path,
                 remote_path,
             )
-        elif action_name == 'create':
+        elif action_name == 'create' and action_name not in skip_actions:
             yield from borgmatic.actions.create.run_create(
                 config_filename,
                 repository,
-                location,
-                storage,
-                hooks,
+                config,
+                config_paths,
                 hook_context,
                 local_borg_version,
                 action_arguments,
@@ -315,13 +337,11 @@ def run_actions(
                 local_path,
                 remote_path,
             )
-        elif action_name == 'prune':
+        elif action_name == 'prune' and action_name not in skip_actions:
             borgmatic.actions.prune.run_prune(
                 config_filename,
                 repository,
-                storage,
-                retention,
-                hooks,
+                config,
                 hook_context,
                 local_borg_version,
                 action_arguments,
@@ -330,13 +350,11 @@ def run_actions(
                 local_path,
                 remote_path,
             )
-        elif action_name == 'compact':
+        elif action_name == 'compact' and action_name not in skip_actions:
             borgmatic.actions.compact.run_compact(
                 config_filename,
                 repository,
-                storage,
-                retention,
-                hooks,
+                config,
                 hook_context,
                 local_borg_version,
                 action_arguments,
@@ -345,15 +363,12 @@ def run_actions(
                 local_path,
                 remote_path,
             )
-        elif action_name == 'check':
-            if checks.repository_enabled_for_checks(repository, consistency):
+        elif action_name == 'check' and action_name not in skip_actions:
+            if checks.repository_enabled_for_checks(repository, config):
                 borgmatic.actions.check.run_check(
                     config_filename,
                     repository,
-                    location,
-                    storage,
-                    consistency,
-                    hooks,
+                    config,
                     hook_context,
                     local_borg_version,
                     action_arguments,
@@ -361,13 +376,11 @@ def run_actions(
                     local_path,
                     remote_path,
                 )
-        elif action_name == 'extract':
+        elif action_name == 'extract' and action_name not in skip_actions:
             borgmatic.actions.extract.run_extract(
                 config_filename,
                 repository,
-                location,
-                storage,
-                hooks,
+                config,
                 hook_context,
                 local_borg_version,
                 action_arguments,
@@ -375,70 +388,110 @@ def run_actions(
                 local_path,
                 remote_path,
             )
-        elif action_name == 'export-tar':
+        elif action_name == 'export-tar' and action_name not in skip_actions:
             borgmatic.actions.export_tar.run_export_tar(
                 repository,
-                storage,
+                config,
                 local_borg_version,
                 action_arguments,
                 global_arguments,
                 local_path,
                 remote_path,
             )
-        elif action_name == 'mount':
+        elif action_name == 'mount' and action_name not in skip_actions:
             borgmatic.actions.mount.run_mount(
                 repository,
-                storage,
-                local_borg_version,
-                arguments['mount'],
-                local_path,
-                remote_path,
-            )
-        elif action_name == 'restore':
-            borgmatic.actions.restore.run_restore(
-                repository,
-                location,
-                storage,
-                hooks,
+                config,
                 local_borg_version,
                 action_arguments,
                 global_arguments,
                 local_path,
                 remote_path,
             )
-        elif action_name == 'rlist':
-            yield from borgmatic.actions.rlist.run_rlist(
-                repository, storage, local_borg_version, action_arguments, local_path, remote_path,
-            )
-        elif action_name == 'list':
-            yield from borgmatic.actions.list.run_list(
-                repository, storage, local_borg_version, action_arguments, local_path, remote_path,
-            )
-        elif action_name == 'rinfo':
-            yield from borgmatic.actions.rinfo.run_rinfo(
-                repository, storage, local_borg_version, action_arguments, local_path, remote_path,
-            )
-        elif action_name == 'info':
-            yield from borgmatic.actions.info.run_info(
-                repository, storage, local_borg_version, action_arguments, local_path, remote_path,
-            )
-        elif action_name == 'break-lock':
-            borgmatic.actions.break_lock.run_break_lock(
+        elif action_name == 'restore' and action_name not in skip_actions:
+            borgmatic.actions.restore.run_restore(
                 repository,
-                storage,
+                config,
                 local_borg_version,
-                arguments['break-lock'],
+                action_arguments,
+                global_arguments,
                 local_path,
                 remote_path,
             )
-        elif action_name == 'borg':
+        elif action_name == 'rlist' and action_name not in skip_actions:
+            yield from borgmatic.actions.rlist.run_rlist(
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
+            )
+        elif action_name == 'list' and action_name not in skip_actions:
+            yield from borgmatic.actions.list.run_list(
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
+            )
+        elif action_name == 'rinfo' and action_name not in skip_actions:
+            yield from borgmatic.actions.rinfo.run_rinfo(
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
+            )
+        elif action_name == 'info' and action_name not in skip_actions:
+            yield from borgmatic.actions.info.run_info(
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
+            )
+        elif action_name == 'break-lock' and action_name not in skip_actions:
+            borgmatic.actions.break_lock.run_break_lock(
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
+            )
+        elif action_name == 'export' and action_name not in skip_actions:
+            borgmatic.actions.export_key.run_export_key(
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
+            )
+        elif action_name == 'borg' and action_name not in skip_actions:
             borgmatic.actions.borg.run_borg(
-                repository, storage, local_borg_version, action_arguments, local_path, remote_path,
+                repository,
+                config,
+                local_borg_version,
+                action_arguments,
+                global_arguments,
+                local_path,
+                remote_path,
             )
 
     command.execute_hook(
-        hooks.get('after_actions'),
-        hooks.get('umask'),
+        config.get('after_actions'),
+        config.get('umask'),
         config_filename,
         'post-actions',
         global_arguments.dry_run,
@@ -450,18 +503,35 @@ def load_configurations(config_filenames, overrides=None, resolve_env=True):
     '''
     Given a sequence of configuration filenames, load and validate each configuration file. Return
     the results as a tuple of: dict of configuration filename to corresponding parsed configuration,
-    and sequence of logging.LogRecord instances containing any parse errors.
+    a sequence of paths for all loaded configuration files (including includes), and a sequence of
+    logging.LogRecord instances containing any parse errors.
+
+    Log records are returned here instead of being logged directly because logging isn't yet
+    initialized at this point!
     '''
     # Dict mapping from config filename to corresponding parsed config dict.
     configs = collections.OrderedDict()
+    config_paths = set()
     logs = []
 
     # Parse and load each configuration file.
     for config_filename in config_filenames:
+        logs.extend(
+            [
+                logging.makeLogRecord(
+                    dict(
+                        levelno=logging.DEBUG,
+                        levelname='DEBUG',
+                        msg=f'{config_filename}: Loading configuration file',
+                    )
+                ),
+            ]
+        )
         try:
-            configs[config_filename], parse_logs = validate.parse_configuration(
+            configs[config_filename], paths, parse_logs = validate.parse_configuration(
                 config_filename, validate.schema_filename(), overrides, resolve_env
             )
+            config_paths.update(paths)
             logs.extend(parse_logs)
         except PermissionError:
             logs.extend(
@@ -491,7 +561,7 @@ def load_configurations(config_filenames, overrides=None, resolve_env=True):
                 ]
             )
 
-    return (configs, logs)
+    return (configs, sorted(config_paths), logs)
 
 
 def log_record(suppress_log=False, **kwargs):
@@ -528,14 +598,24 @@ def log_error_records(
         raise error
     except CalledProcessError as error:
         yield log_record(levelno=levelno, levelname=level_name, msg=message)
+
         if error.output:
-            # Suppress these logs for now and save full error output for the log summary at the end.
-            yield log_record(
-                levelno=levelno,
-                levelname=level_name,
-                msg=error.output,
-                suppress_log=not log_command_error_output,
-            )
+            try:
+                output = error.output.decode('utf-8')
+            except (UnicodeDecodeError, AttributeError):
+                output = error.output
+
+            # Suppress these logs for now and save the error output for the log summary at the end.
+            # Log a separate record per line, as some errors can be really verbose and overflow the
+            # per-record size limits imposed by some logging backends.
+            for output_line in output.splitlines():
+                yield log_record(
+                    levelno=levelno,
+                    levelname=level_name,
+                    msg=output_line,
+                    suppress_log=True,
+                )
+
         yield log_record(levelno=levelno, levelname=level_name, msg=error)
     except (ValueError, OSError) as error:
         yield log_record(levelno=levelno, levelname=level_name, msg=message)
@@ -551,15 +631,109 @@ def get_local_path(configs):
     Arbitrarily return the local path from the first configuration dict. Default to "borg" if not
     set.
     '''
-    return next(iter(configs.values())).get('location', {}).get('local_path', 'borg')
+    return next(iter(configs.values())).get('local_path', 'borg')
 
 
-def collect_configuration_run_summary_logs(configs, arguments):
+def collect_highlander_action_summary_logs(configs, arguments, configuration_parse_errors):
     '''
-    Given a dict of configuration filename to corresponding parsed configuration, and parsed
-    command-line arguments as a dict from subparser name to a parsed namespace of arguments, run
-    each configuration file and yield a series of logging.LogRecord instances containing summary
-    information about each run.
+    Given a dict of configuration filename to corresponding parsed configuration, parsed
+    command-line arguments as a dict from subparser name to a parsed namespace of arguments, and
+    whether any configuration files encountered errors during parsing, run a highlander action
+    specified in the arguments, if any, and yield a series of logging.LogRecord instances containing
+    summary information.
+
+    A highlander action is an action that cannot coexist with other actions on the borgmatic
+    command-line, and borgmatic exits after processing such an action.
+    '''
+    add_custom_log_levels()
+
+    if 'bootstrap' in arguments:
+        try:
+            # No configuration file is needed for bootstrap.
+            local_borg_version = borg_version.local_borg_version({}, 'borg')
+        except (OSError, CalledProcessError, ValueError) as error:
+            yield from log_error_records('Error getting local Borg version', error)
+            return
+
+        try:
+            borgmatic.actions.config.bootstrap.run_bootstrap(
+                arguments['bootstrap'], arguments['global'], local_borg_version
+            )
+            yield logging.makeLogRecord(
+                dict(
+                    levelno=logging.ANSWER,
+                    levelname='ANSWER',
+                    msg='Bootstrap successful',
+                )
+            )
+        except (
+            CalledProcessError,
+            ValueError,
+            OSError,
+        ) as error:
+            yield from log_error_records(error)
+
+        return
+
+    if 'generate' in arguments:
+        try:
+            borgmatic.actions.config.generate.run_generate(
+                arguments['generate'], arguments['global']
+            )
+            yield logging.makeLogRecord(
+                dict(
+                    levelno=logging.ANSWER,
+                    levelname='ANSWER',
+                    msg='Generate successful',
+                )
+            )
+        except (
+            CalledProcessError,
+            ValueError,
+            OSError,
+        ) as error:
+            yield from log_error_records(error)
+
+        return
+
+    if 'validate' in arguments:
+        if configuration_parse_errors:
+            yield logging.makeLogRecord(
+                dict(
+                    levelno=logging.CRITICAL,
+                    levelname='CRITICAL',
+                    msg='Configuration validation failed',
+                )
+            )
+
+            return
+
+        try:
+            borgmatic.actions.config.validate.run_validate(arguments['validate'], configs)
+
+            yield logging.makeLogRecord(
+                dict(
+                    levelno=logging.ANSWER,
+                    levelname='ANSWER',
+                    msg='All configuration files are valid',
+                )
+            )
+        except (
+            CalledProcessError,
+            ValueError,
+            OSError,
+        ) as error:
+            yield from log_error_records(error)
+
+        return
+
+
+def collect_configuration_run_summary_logs(configs, config_paths, arguments):
+    '''
+    Given a dict of configuration filename to corresponding parsed configuration, a sequence of
+    loaded configuration paths, and parsed command-line arguments as a dict from subparser name to a
+    parsed namespace of arguments, run each configuration file and yield a series of
+    logging.LogRecord instances containing summary information about each run.
 
     As a side effect of running through these configuration files, output their JSON results, if
     any, to stdout.
@@ -590,10 +764,9 @@ def collect_configuration_run_summary_logs(configs, arguments):
     if 'create' in arguments:
         try:
             for config_filename, config in configs.items():
-                hooks = config.get('hooks', {})
                 command.execute_hook(
-                    hooks.get('before_everything'),
-                    hooks.get('umask'),
+                    config.get('before_everything'),
+                    config.get('umask'),
                     config_filename,
                     'pre-everything',
                     arguments['global'].dry_run,
@@ -605,7 +778,7 @@ def collect_configuration_run_summary_logs(configs, arguments):
     # Execute the actions corresponding to each configuration file.
     json_results = []
     for config_filename, config in configs.items():
-        results = list(run_configuration(config_filename, config, arguments))
+        results = list(run_configuration(config_filename, config, config_paths, arguments))
         error_logs = tuple(result for result in results if isinstance(result, logging.LogRecord))
 
         if error_logs:
@@ -626,7 +799,9 @@ def collect_configuration_run_summary_logs(configs, arguments):
         logger.info(f"Unmounting mount point {arguments['umount'].mount_point}")
         try:
             borg_umount.unmount_archive(
-                mount_point=arguments['umount'].mount_point, local_path=get_local_path(configs),
+                config,
+                mount_point=arguments['umount'].mount_point,
+                local_path=get_local_path(configs),
             )
         except (CalledProcessError, OSError) as error:
             yield from log_error_records('Error unmounting mount point', error)
@@ -637,10 +812,9 @@ def collect_configuration_run_summary_logs(configs, arguments):
     if 'create' in arguments:
         try:
             for config_filename, config in configs.items():
-                hooks = config.get('hooks', {})
                 command.execute_hook(
-                    hooks.get('after_everything'),
-                    hooks.get('umask'),
+                    config.get('after_everything'),
+                    config.get('umask'),
                     config_filename,
                     'post-everything',
                     arguments['global'].dry_run,
@@ -658,7 +832,7 @@ def exit_with_help_link():  # pragma: no cover
     sys.exit(1)
 
 
-def main():  # pragma: no cover
+def main(extra_summary_logs=[]):  # pragma: no cover
     configure_signals()
 
     try:
@@ -676,24 +850,28 @@ def main():  # pragma: no cover
 
     global_arguments = arguments['global']
     if global_arguments.version:
-        print(pkg_resources.require('borgmatic')[0].version)
+        print(importlib.metadata.version('borgmatic'))
         sys.exit(0)
     if global_arguments.bash_completion:
-        print(borgmatic.commands.completion.bash_completion())
+        print(borgmatic.commands.completion.bash.bash_completion())
+        sys.exit(0)
+    if global_arguments.fish_completion:
+        print(borgmatic.commands.completion.fish.fish_completion())
         sys.exit(0)
 
     config_filenames = tuple(collect.collect_config_filenames(global_arguments.config_paths))
-    configs, parse_logs = load_configurations(
+    configs, config_paths, parse_logs = load_configurations(
         config_filenames, global_arguments.overrides, global_arguments.resolve_env
     )
+    configuration_parse_errors = (
+        (max(log.levelno for log in parse_logs) >= logging.CRITICAL) if parse_logs else False
+    )
 
     any_json_flags = any(
         getattr(sub_arguments, 'json', False) for sub_arguments in arguments.values()
     )
-    colorama.init(
-        autoreset=True,
-        strip=not should_do_markup(global_arguments.no_color or any_json_flags, configs),
-    )
+    color_enabled = should_do_markup(global_arguments.no_color or any_json_flags, configs)
+    colorama.init(autoreset=color_enabled, strip=not color_enabled)
     try:
         configure_logging(
             verbosity_to_log_level(global_arguments.verbosity),
@@ -702,16 +880,25 @@ def main():  # pragma: no cover
             verbosity_to_log_level(global_arguments.monitoring_verbosity),
             global_arguments.log_file,
             global_arguments.log_file_format,
+            color_enabled=color_enabled,
         )
     except (FileNotFoundError, PermissionError) as error:
         configure_logging(logging.CRITICAL)
         logger.critical(f'Error configuring logging: {error}')
         exit_with_help_link()
 
-    logger.debug('Ensuring legacy configuration is upgraded')
-    convert.guard_configuration_upgraded(LEGACY_CONFIG_PATH, config_filenames)
-
-    summary_logs = parse_logs + list(collect_configuration_run_summary_logs(configs, arguments))
+    summary_logs = (
+        extra_summary_logs
+        + parse_logs
+        + (
+            list(
+                collect_highlander_action_summary_logs(
+                    configs, arguments, configuration_parse_errors
+                )
+            )
+            or list(collect_configuration_run_summary_logs(configs, config_paths, arguments))
+        )
+    )
     summary_logs_max_level = max(log.levelno for log in summary_logs)
 
     for message in ('', 'summary:'):
